import pandas as pd

from .data import data_path, resource

__all__ = ["load"]


def load():
    _fetch()

    name_corr = {"Xloc": "x", "Yloc": "y", "Landuse": "land", "Rock": "rock"}

    # Load training data.
    train = pd.read_csv(data_path("jura", "prediction.dat"), sep=r"\s+")
    train.columns = [name_corr[c] if c in name_corr else c for c in train.columns]
    train.set_index(["x", "y"], inplace=True)

    # Load test data.
    test = pd.read_csv(data_path("jura", "validation.dat"), sep=r"\s+")
    test.columns = [name_corr[c] if c in name_corr else c for c in test.columns]
    test.set_index(["x", "y"], inplace=True)

    # Setup according to experiment.
    train = pd.concat([train[["Ni", "Zn", "Cd"]], test[["Ni", "Zn"]]])
    test = test[["Cd"]]

    return train, test


def _fetch():
    resource(
        target=data_path("jura", "prediction.dat"),
        url="https://www.dropbox.com/s/n7jamwpsi9p4wvs/prediction.dat?dl=1",
    )
    resource(
        target=data_path("jura", "validation.dat"),
        url="https://www.dropbox.com/s/slvclnw8qr6gjv8/validation.dat?dl=1",
    )



import inspect
import warnings
from functools import reduce
from operator import mul

import lab as B

__all__ = ["indented_kv", "warn_upmodule", "inv_perm", "normal1d_logpdf", "BatchVars"]


def indented_kv(key: str, value: str, indent=1, separator="=", suffix=""):
    """Print something as a key-value pair whilst properly indenting. This is useful
    for implementations of`str` and `repr`.

    Args:
        key (str): Key.
        value (str): Value.
        indent (int, optional): Number of spaces to indent. Defaults to 1.
        separator (str, optional): Separator between the key and value. Defaults to "=".
        suffix (str, optional): Extra to print at the end. You can set this, e.g., to
            ",\n" or ">". Defaults to no suffix.

    Returns
        str: Key-value representation with proper indentation.
    """
    key_string = f"{indent * ' '}{key}{separator}"
    value_string = value.strip().replace("\n", "\n" + " " * len(key_string))
    return key_string + value_string + suffix


def _get_module_name(frame_info):
    return frame_info.frame.f_globals["__name__"].split(".")[0]


def warn_upmodule(*args, **kw_args):
    """Call :func:`warnings.warn`, but set the keyword argument `stacklevel`
    corresponding to the first other module above the current module.
    """
    level = 2
    stack = inspect.stack()[1:]
    this_module = _get_module_name(stack[0])

    # Up the level until the first other module is reached.
    for frame_info in stack:
        if _get_module_name(frame_info) != this_module:
            break
        else:
            level += 1

    warnings.warn(*args, stacklevel=level, **kw_args)


def inv_perm(perm):
    """Invert a permutation.

    Args:
        perm (list): Permutation to invert.

    Returns:
        list: Inverse permutation.
    """
    out = [0] * len(perm)
    for i, p in enumerate(perm):
        out[p] = i
    return out


def normal1d_logpdf(x, var, mean=0):
    """Broadcast the one-dimensional normal logpdf.

    Args:
        x (tensor): Point to evaluate at.
        var (tensor): Variances.
        mean (tensor): Means.

    Returns:
        tensor: Logpdf.
    """
    return -(B.log_2_pi + B.log(var) + (x - mean) ** 2 / var) / 2


class BatchVars:
    """Extract variables from a source with a batch axis as the first axis.

    Args:
        source (tensor): Source.
    """

    def __init__(self, source):
        self.source = source
        self.index = 0

    def get(self, shape):
        """Get a batch of tensor of a particular shape.

        Args:
            shape (shape): Shape of tensor.

        Returns:
            tensor: Batch of tensors of shape `shape`.
        """
        length = reduce(mul, shape, 1)
        res = self.source[:, self.index : self.index + length]
        self.index += length
        return B.reshape(res, -1, *shape)



import datetime

import matplotlib.pyplot as plt
import numpy as np
import pandas

from .data import data_path, asserted_dependency
from .. import out
from ..plot import tweak

__all__ = ["load", "stats_and_vis"]


def load():
    _fetch()

    data = pandas.read_csv(data_path("station", "global_station_data.csv"))
    data.rename(columns={"value": "temp"}, inplace=True)

    # Extract locations per year and month.
    per_month = []
    data.groupby(["year", "month"]).apply(lambda x: per_month.append(x.copy()))

    # Extract dates.
    per_month_dates = [
        datetime.date(year=x["year"].iloc[0], month=x["month"].iloc[0], day=1)
        for x in per_month
    ]

    return data, per_month_dates, per_month


def _fetch():
    asserted_dependency(target=data_path("station", "global_station_data.csv"))


def stats_and_vis():
    data, per_month_dates, per_month = load()

    # Print some stats.
    num_measurements = np.array(list(map(len, per_month)))
    out.kv("Total number of dates", len(per_month))
    out.kv(
        "Number of dates with 20 measurements or more", np.sum(num_measurements >= 20)
    )
    out.kv("Average number of measurements", int(np.mean(num_measurements)))

    # Show number of measurements.
    plt.figure()
    plt.title("Number Of Measurements")
    plt.scatter(per_month_dates, num_measurements, s=2)
    plt.xlabel("Date")
    plt.ylabel("Number of measurements")
    tweak(legend=False)

    # Show 16 randomly chosen snapshots.
    plt.figure(figsize=(12, 8))
    vmin = data["temp"].min()
    vmax = data["temp"].max()
    lon_range = data["lon"].min(), data["lon"].max()
    lat_range = data["lat"].min(), data["lat"].max()
    inds = sorted(np.random.permutation(len(per_month))[:16])
    for i, ind in enumerate(inds):
        plt.subplot(4, 4, i + 1).set_title(per_month_dates[ind])
        plt.scatter(
            per_month[ind]["lon"],
            per_month[ind]["lat"],
            s=4,
            c=per_month[ind]["temp"],
            vmin=vmin,
            vmax=vmax,
        )
        plt.colorbar()
        plt.xlim(*lon_range)
        plt.ylim(*lat_range)
        tweak(legend=False)

    plt.show()



import pytest

import lab as B
import numpy as np
import tensorflow as tf
import torch
import time

import wbml.out as out
from .util import approx


class RecordingStream:
    """A stream that records its writes."""

    def __init__(self):
        self.writes = []

    def write(self, msg):
        self.writes.append(msg)

    def flush(self):
        self.writes.append("[flush]")

    def __str__(self):
        return "".join(self.writes)

    def __len__(self):
        return len(self.writes)


class Mock:
    """Mock the stream that `wbml.out` uses. Also reset
    `wbml.out._time_start`."""

    def __enter__(self):
        self.stream = RecordingStream()

        # Save current stream and time started.
        self.saved_streams = list(out.streams)
        self.saved_time_start = out._time_start

        # Mock stream and time started.
        out.streams = [self.stream]
        out._time_start = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Unmock stream and time started.
        out.streams = self.saved_streams
        out._time_start = self.saved_time_start

    def __str__(self):
        return str(self.stream)

    def __repr__(self):
        return "<Mock: recorded=" + str(self.stream) + ">"

    def __len__(self):
        return len(self.stream)

    def __getitem__(self, i):
        return self.stream.writes[i]


def test_section():
    with Mock() as mock:
        out.out("before")

        with out.Section():
            out.out("message1")

        with out.Section("name"):
            out.out("message2")

            with out.Section():
                out.out("message3")

        out.out("after")

    assert len(mock) == 6
    assert mock[0] == "before\n"
    assert mock[1] == "    message1\n"
    assert mock[2] == "name:\n"
    assert mock[3] == "    message2\n"
    assert mock[4] == "        message3\n"
    assert mock[5] == "after\n"


def test_out():
    with Mock() as mock:
        out.out("message")

    assert len(mock) == 1
    assert str(mock) == "message\n"


def test_out_newlines():
    # Test that newlines are correctly indented.
    with Mock() as mock:
        out.out("a\nb")

        with out.Section():
            out.out("c\nd")

    assert len(mock) == 2
    assert mock[0] == "a\nb\n"
    assert mock[1] == "    c\n    d\n"


def test_kv():
    with Mock() as mock:
        out.kv("key", "value")
        out.kv(1.0, 1.0)
        out.kv(10.0, 10.0)
        out.kv(1000.0, 1000.0)
        out.kv(1000, 1000)

    assert len(mock) == 5
    assert mock[0] == "key:        value\n"
    assert mock[1] == "1.0:        1.0\n"
    assert mock[2] == "10.0:       10.0\n"
    assert mock[3] == "1.000e+03:  1.000e+03\n"
    assert mock[4] == "1000:       1000\n"


def test_kv_unit():
    with Mock() as mock:
        out.kv("key", 1, unit="s")

    assert len(mock) == 1
    assert mock[0] == "key:        1 s\n"


def test_kv_fmt():
    with Mock() as mock:
        out.kv("key", 1, fmt=".4f")

    assert len(mock) == 1
    assert mock[0] == "key:        1.0000\n"


def test_kv_dict():
    # Test giving a dictionary.
    with Mock() as mock:
        out.kv({"level1": {"level2": {1: 1}}})

    assert len(mock) == 3
    assert mock[0] == "level1:\n"
    assert mock[1] == "    level2:\n"
    assert mock[2] == "        1:          1\n"


def test_kv_dict_as_value():
    # Test giving a key and a dictionary.
    with Mock() as mock:
        out.kv("dict", {1: 1})

    assert len(mock) == 2
    assert mock[0] == "dict:\n"
    assert mock[1] == "    1:          1\n"


def test_kv_newlines():
    # Test values with newlines.
    with Mock() as mock:
        out.kv("a", "b\nc")

    assert len(mock) == 2
    assert mock[0] == "a:\n"
    assert mock[1] == "    b\n    c\n"


def test_format_object():
    class A:
        def __str__(self):
            return "FormattedA()"

    assert out.format(A()) == "FormattedA()"


@pytest.mark.parametrize(
    "x, y",
    [
        # Floats:
        (0.000012, "1.200e-05"),
        (0.00012, "1.200e-04"),
        (0.0012, "1.200e-03"),
        (0.012, "0.012"),
        (0.12, "0.12"),
        (1.2, "1.2"),
        (12.0, "12.0"),
        (120.0, "120.0"),
        (1200.0, "1.200e+03"),
        (12000.0, "1.200e+04"),
        # Integers:
        (1, "1"),
        (1200, "1200"),
        (12000, "12000"),
        # NaN and infinity:
        (np.nan, "nan"),
        (np.inf, "inf"),
        (-np.inf, "-inf"),
        # Containers:
        ([1, 1.0, 1000.0], "[1, 1.0, 1.000e+03]"),
        ((1, 1.0, 1000.0), "(1, 1.0, 1.000e+03)"),
        ({1}, "{1}"),
        ({1.0}, "{1.0}"),
        ({1000.0}, "{1.000e+03}"),
        # NumPy:
        (np.array(0.000012), "1.200e-05"),
        (B.ones(int, 3), "[1 1 1]"),
        (
            B.ones(int, 3, 3),
            "(3x3 array of data type int64)\n[[1 1 1]\n [1 1 1]\n [1 1 1]]",
        ),
        # PyTorch:
        (B.ones(torch.int, 3), "[1 1 1]"),
        # TensorFlow:
        (B.ones(tf.int32, 3), "[1 1 1]"),
    ],
)
def test_format(x, y):
    assert out.format(x) == y


def test_format_info_flag():
    assert out.format(B.ones(int, 3, 3), False) == "[[1 1 1]\n [1 1 1]\n [1 1 1]]"


def _assert_counts(mock):
    assert len(mock) == 8
    assert mock[1] == "[flush]"
    assert mock[2] == " 1"
    assert mock[3] == "[flush]"
    assert mock[4] == " 2"
    assert mock[5] == "[flush]"
    assert mock[6] == "\n"
    assert mock[7] == "[flush]"


def test_counter():
    # Test normal application.
    with Mock() as mock:
        with out.Counter() as counter:
            counter.count()
            counter.count()

    assert mock[0] == "Counting:"
    _assert_counts(mock)


def test_counter_total():
    with Mock() as mock:
        with out.Counter(name="name", total=3) as counter:
            counter.count()
            counter.count()

    assert mock[0] == "name (total: 3):"
    _assert_counts(mock)


def test_counter_map():
    # Test mapping.
    with Mock() as mock:
        res = out.Counter.map(lambda x: x ** 2, [2, 3])

    assert res == [4, 9]
    assert mock[0] == "Mapping (total: 2):"
    _assert_counts(mock)


def test_counter_map2():
    with Mock() as mock:
        res = out.Counter.map(lambda x: x ** 2, [2, 3], name="name")

    assert res == [4, 9]
    assert mock[0] == "name (total: 2):"
    _assert_counts(mock)


def test_compute_alpha():
    lags = 8
    alpha = out._compute_alpha(lags)
    x = np.sin(2 * np.pi / lags * B.range(10000))

    # Perform filtering.
    y = [x[0]]
    for xi in x:
        y.append(alpha * xi + (1 - alpha) * y[-1])
    y = np.array(y)

    # Check damping in dB.
    ratio = 10 * np.log10(np.mean(y ** 2) / np.mean(x ** 2))
    approx(ratio, -3, atol=5e-2)

    # Check not setting the cut-off.
    assert out._compute_alpha(None) == 1


def test_progress():
    # Test a simple case.
    with Mock() as mock:
        with out.Progress() as progress:
            progress(a=1)

    assert len(mock) == 5
    assert mock[0] == "Progress:\n"
    assert mock[1] == "    Iteration 1:\n"
    assert "Time elapsed" in mock[2]
    assert mock[3] == "        a:          1\n"
    assert mock[4] == "    Done!\n"


def test_progress_interval():
    # Test setting the total number of iterations and report interval as int.
    with Mock() as mock:
        with out.Progress(name="name", total=4, interval=3) as progress:
            progress(a="a")
            progress(a="b")
            progress(a="c")
            # Change time per iteration to a second to show a non-zero time
            # left.
            progress.values["_time_per_it"] = 1
            progress(a="d")

    assert len(mock) == 14
    assert mock[0] == "name:\n"
    # First is always shown.
    assert mock[1] == "    Iteration 1/4:\n"
    assert "Time elapsed" in mock[2]
    assert "Time left" in mock[3]
    assert mock[4] == "        a:          a\n"
    assert mock[5] == "    Iteration 3/4:\n"
    assert "Time elapsed" in mock[6]
    assert "Time left" in mock[7]
    assert mock[8] == "        a:          c\n"
    # Last is also always shown.
    assert mock[9] == "    Iteration 4/4:\n"
    assert "Time elapsed" in mock[10]
    assert "Time left" in mock[11]
    assert "0.0 s" not in mock[11]
    assert mock[12] == "        a:          d\n"
    assert mock[13] == "    Done!\n"


def test_progress_filters():
    # Test filters, report interval as float, and giving a dictionary.
    with Mock() as mock:
        with out.Progress(
            name="name", interval=1e-10, filter={"a": None}, filter_global=np.inf
        ) as progress:
            progress({"a": 1, "b": 1})
            progress({"a": 2, "b": 2})

    assert len(mock) == 10
    assert mock[0] == "name:\n"
    assert mock[1] == "    Iteration 1:\n"
    assert "Time elapsed" in mock[2]
    assert mock[3] == "        a:          1\n"
    assert mock[4] == "        b:          1\n"
    assert mock[5] == "    Iteration 2:\n"
    assert "Time elapsed" in mock[6]
    # Filter should be off.
    assert mock[7] == "        a:          2\n"
    # Filter should be maximal.
    assert mock[8] == "        b:          1.0\n"
    assert mock[9] == "    Done!\n"


def test_progress_map():
    with Mock() as mock:
        res = out.Progress.map(lambda x: x ** 2, [2, 3])

    assert res == [4, 9]
    assert len(mock) == 8
    assert mock[0] == "Mapping:\n"


def test_progress_map2():
    with Mock() as mock:
        res = out.Progress.map(lambda x: x ** 2, [2, 3], name="name")

    assert res == [4, 9]
    assert len(mock) == 8
    assert mock[0] == "name:\n"


def test_time_report_interval(monkeypatch):
    monkeypatch.setattr(out, "report_time", True)

    # Test that time stamp is not repeated unnecessarily.
    with Mock() as mock:
        out.out("a")
        out.out("b")
        time.sleep(1.0)
        out.out("c")

    assert len(mock) == 3
    assert mock[0] == "00:00:00 | a\n"
    assert mock[1] == "         | b\n"
    assert mock[2] == "00:00:01 | c\n"


def test_time_report_calculation(monkeypatch):
    monkeypatch.setattr(out, "report_time", True)

    # Test that time is correctly calculated.
    with Mock() as mock:
        out._time_start = time.time() - 2 * 60 * 60 - 2 * 60 - 2
        out.out("a")

    assert len(mock) == 1
    assert mock[0] == "02:02:02 | a\n"


def test_time_report_with_progress(monkeypatch):
    monkeypatch.setattr(out, "report_time", True)

    # Test normal application of counting, as above.
    with Mock() as mock:
        with out.Counter() as counter:
            counter.count()
            counter.count()

    assert len(mock) == 8
    assert mock[0] == "00:00:00 | Counting:"
    assert mock[1] == "[flush]"
    assert mock[2] == " 1"
    assert mock[3] == "[flush]"
    assert mock[4] == " 2"
    assert mock[5] == "[flush]"
    assert mock[6] == "\n"
    assert mock[7] == "[flush]"



import inspect
import warnings

__all__ = ["warn_upmodule"]

# By default show deprecation warnings.
warnings.filterwarnings("always", category=DeprecationWarning)


def _get_module_name(frame_info):
    return frame_info.frame.f_globals["__name__"].split(".")[0]


def warn_upmodule(*args, **kw_args):
    """Call :func:`warnings.warn`, but set the keyword argument `stacklevel`
    corresponding to the first other module above the current module.
    """
    level = 2
    stack = inspect.stack()[1:]
    this_module = _get_module_name(stack[0])

    # Up the level until the first other module is reached.
    for frame_info in stack:
        if _get_module_name(frame_info) != this_module:
            break
        else:
            level += 1

    warnings.warn(*args, stacklevel=level, **kw_args)



import numpy as np
import pandas as pd
import scipy.io

from .data import data_path, resource, dependency

__all__ = ["load"]


def load():
    _fetch()

    # Compute angles.
    azimuths = np.concatenate(
        (
            np.array([-80, -65, -55]),
            np.arange(-45, 45 + 1, 5),
            np.array([55, 65, 80]),
        )
    )[:, None]
    elevations = (-45 + 5.625 * np.arange(50))[None, :]
    azimuths, elevations = np.broadcast_arrays(azimuths, elevations)

    # Load data.
    data = scipy.io.loadmat(
        data_path("kemar", "standard_kemar", "subject_021", "hrir_final.mat")
    )
    left = data["hrir_l"]
    right = data["hrir_r"]

    # Compute time.
    t = np.arange(left.shape[2]) / 441000

    # Arrange into convenient data frames.
    angles = pd.MultiIndex.from_frame(
        pd.DataFrame({"azimuth": azimuths.flatten(), "elevation": elevations.flatten()})
    )
    left = pd.DataFrame(left.reshape(len(angles), -1).T, columns=angles, index=t)
    right = pd.DataFrame(right.reshape(len(angles), -1).T, columns=angles, index=t)

    return {"left": left, "right": right}


def _fetch():
    resource(
        target=data_path("kemar", "standard_kemar.tar"),
        url=(
            "https://ucdavis.app.box.com/index.php"
            "?rm=box_download_shared_file"
            "&shared_name=0418n9rcvh2krxg7zy4prbgkam90a56a"
            "&file_id=f_245793603750"
        ),
    )
    dependency(
        target=data_path("kemar", "standard_kemar"),
        source=data_path("kemar", "standard_kemar.tar"),
        commands=[
            "mkdir standard_kemar",
            "tar xzf standard_kemar.tar -C standard_kemar",
        ],
    )



import datetime
import os
import shutil
import subprocess
import urllib.request
from contextlib import closing

import numpy as np
import pandas as pd
import requests

import wbml.out

__all__ = [
    "DependencyError",
    "resource",
    "dependency",
    "asserted_dependency",
    "split_df",
    "data_path",
    "date_to_decimal_year",
]


class DependencyError(AssertionError):
    """Exception raised in case of an erroneous dependency."""


def resource(target, url, post=False, **kw_args):
    """Specify a dependency on an online resource.

    Further takes in keyword arguments that are passed to the appropriate method
    from :mod:`requests` or :mod:`urllib`.

    Args:
        target (str): Target file.
        url (str): Source URL.
        post (bool, optional): Make a POST request instead of a GET request.
            Only applicable if the URL starts with "http" or "https". Defaults
            to `False`.
    """
    if not os.path.exists(target):
        with wbml.out.Section("Downloading file"):
            wbml.out.kv("Source", url)
            wbml.out.kv("Target", target)

            # Ensure that all directories in the path exist.
            make_dirs(target)

            # If the URL starts with "ftp", use the :mod:`urllib` library.
            if url.startswith("ftp"):
                with closing(urllib.request.urlopen(url, **kw_args)) as r:
                    with open(target, "wb") as f:
                        shutil.copyfileobj(r, f)

            # By default, use the :mod:`requests` library.
            else:
                request = requests.post if post else requests.get
                with request(url, stream=True, **kw_args) as r:
                    with open(target, "wb") as f:
                        shutil.copyfileobj(r.raw, f)


def dependency(target, source, commands):
    """Specify a dependency that is generated from an existing file.

    Args:
        target (str): Target file.
        source (str): Source file.
        commands (list[str]): List of commands to generate target file.
    """
    if not os.path.exists(target):
        with wbml.out.Section("Generating file"):
            wbml.out.kv("Source", source)
            wbml.out.kv("Target", target)

            # Check that the source exists.
            if not os.path.exists(source):
                raise DependencyError(
                    f'Source "{source}" asserted to exist, but it does not.'
                )

            # Save current working directory.
            current_wd = os.getcwd()

            # Ensure that all directories in the path exist.
            make_dirs(target)

            # Perform commands.
            for command in commands:
                wbml.out.out(command)

                # Change working directory to directory of target file, run
                # command, and restore working directory afterwards.
                os.chdir(os.path.dirname(target))
                subprocess.call(command, shell=True)
                os.chdir(current_wd)


def asserted_dependency(target):
    """Specify a dependency that cannot be fetched.

    Args:
        target (str): Target file.
    """
    if not os.path.exists(target):
        raise DependencyError(
            f'Dependency "{target}" is asserted to exist, '
            f"but it does not, and it cannot be "
            f"automatically fetched. Please put the file "
            f"into place manually."
        )


def make_dirs(path):
    """Make the directories in the path of a file.

    Args:
        path (url): Path of a file.
    """
    os.makedirs(os.path.dirname(path), exist_ok=True)


def data_path(*xs):
    """Get the path of a data file.

    Args:
        *xs (str): Parts of the path.

    Returns:
        str: Absolute path.
    """
    return os.path.abspath(
        os.path.join(os.path.dirname(__file__), os.pardir, os.pardir, "data", *xs)
    )


def split_df(df, index_range, columns, iloc=False):
    """Split a data frame by selecting from columns a particular range.

    Args:
        df (:class:`pd.DataFrame`): Data frame to split.
        index_range (tuple): Tuple containing lower and upper limit of the
            range to split the index by. If `index_range = (a, b)`, then
            `[a, b)` is taken.
        columns (list[object]): Columns to select.
        iloc (bool, optional): The index range is the integer location instead
            of the index value. Defaults to `False`.

    Returns:
        tuple[:class:`pd.DataFrame`]: Selected rows from selected columns
            and the remainder.
    """
    if iloc:
        inds = np.arange(df.shape[0])
        rows = (inds >= index_range[0]) & (inds < index_range[1])
    else:
        rows = (df.index >= index_range[0]) & (df.index < index_range[1])
    selected = pd.DataFrame([df[name][rows] for name in columns]).T
    remainder = pd.DataFrame(
        [df[name][~rows] for name in columns]
        + [df[name] for name in set(df.columns) - set(columns)]
    ).T

    # Fix order of columns.
    selected_inds = [i for i, c in enumerate(df.columns) if c in columns]
    selected = selected.reindex(df.columns[np.array(selected_inds)], axis=1)
    remainder = remainder.reindex(df.columns, axis=1)

    return selected, remainder


def date_to_decimal_year(date, format=None):
    """Convert a date to decimal year.

    Args:
        date (str): Date as a string.
        format (str, optional): Format of the date if a conversion is needed.

    Returns:
        float: Decimal year corresponding to the date.
    """
    if format:
        date = datetime.datetime.strptime(date, format)
    start = datetime.date(date.year, 1, 1).toordinal()
    year_length = datetime.date(date.year + 1, 1, 1).toordinal() - start

    # Account for subday time.
    subday_time = 0
    if hasattr(date, "hour"):
        subday_time += date.hour / year_length / 24
    if hasattr(date, "minute"):
        subday_time += date.minute / year_length / 24 / 60
    if hasattr(date, "second"):
        subday_time += date.second / year_length / 24 / 60 / 60

    return date.year + float(date.toordinal() - start) / year_length + subday_time



import lab as B
import numpy as np
import pytest
from varz import Vars

from wbml.net import Normalise, Linear, Activation, Recurrent, GRU, Elman, ff, rnn
from .util import approx, approx


def check_batch_consistency(layer, xs):
    outs = []
    for x in xs:
        outs.append(layer(x))

    # Test consistency. This tests that rank 2 and 3 tensors are of the correct
    # shape.
    approx(layer(xs), B.stack(*outs, axis=0))


def test_normalise():
    layer = Normalise(epsilon=0)
    x = B.randn(10, 5, 3)

    # Check number of weights and width.
    assert layer.num_weights(10) == 0
    assert layer.width == 10

    # Check initialisation and width.
    layer.initialise(3, None)
    assert layer.width == 3

    # Check correctness
    out = layer(x)
    approx(B.std(out, axis=2), B.ones(10, 5), rtol=1e-4)
    approx(B.mean(out, axis=2), B.zeros(10, 5), atol=1e-4)


def test_linear():
    layer = Linear(20)
    x = B.randn(10, 5, 3)

    # Check number of weights and width.
    assert layer.num_weights(3) == 3 * 20 + 20
    assert layer.width == 20

    # Check initialisation and width.
    vs = Vars(np.float64)
    layer.initialise(3, vs)
    assert layer.width == 20

    # Check batch consistency.
    check_batch_consistency(layer, x)

    # Check correctness.
    approx(layer(x), B.matmul(x, layer.A[None, :, :]) + layer.b[None, :, :])


def test_activation():
    layer = Activation()
    x = B.randn(10, 5, 3)

    # Check number of weights and width.
    assert layer.num_weights(10) == 0
    assert layer.width == 10

    # Check initialisation and width.
    layer.initialise(3, None)
    assert layer.width == 3

    # Check correctness
    approx(layer(x), B.relu(x))


def test_recurrent():
    vs = Vars(np.float32)

    # Test setting the initial hidden state.
    layer = Recurrent(GRU(10), B.zeros(1, 10))
    layer.initialise(5, vs)
    approx(layer.h0, B.zeros(1, 10))

    layer = Recurrent(GRU(10))
    layer.initialise(5, vs)
    assert layer.h0 is not None

    # Check batch consistency.
    check_batch_consistency(layer, B.randn(30, 20, 5))

    # Test preservation of rank upon calls.
    assert B.shape(layer(B.randn(20, 5))) == (20, 10)
    assert B.shape(layer(B.randn(30, 20, 5))) == (30, 20, 10)

    # Check that zero-dimensional calls fail.
    with pytest.raises(ValueError):
        layer(0)


def test_ff():
    vs = Vars(np.float32)

    nn = ff(10, (20, 30), normalise=True)
    nn.initialise(5, vs)
    x = B.randn(2, 3, 5)

    # Check number of weights and width.
    assert B.length(vs.get_vector()) == nn.num_weights(5)
    assert nn.width == 10

    # Test batch consistency.
    check_batch_consistency(nn, x)

    # Check composition.
    assert len(nn.layers) == 7
    assert type(nn.layers[0]) == Linear
    assert nn.layers[0].A.shape[0] == 5
    assert nn.layers[0].width == 20
    assert type(nn.layers[1]) == Activation
    assert nn.layers[1].width == 20
    assert type(nn.layers[2]) == Normalise
    assert nn.layers[2].width == 20
    assert type(nn.layers[3]) == Linear
    assert nn.layers[3].width == 30
    assert type(nn.layers[4]) == Activation
    assert nn.layers[4].width == 30
    assert type(nn.layers[5]) == Normalise
    assert nn.layers[5].width == 30
    assert type(nn.layers[6]) == Linear
    assert nn.layers[6].width == 10

    # Check that one-dimensional calls are okay.
    vs = Vars(np.float32)
    nn.initialise(1, vs)
    approx(nn(B.linspace(0, 1, 10)), nn(B.linspace(0, 1, 10)[:, None]))

    # Check that zero-dimensional calls fail.
    with pytest.raises(ValueError):
        nn(0)

    # Check normalisation layers disappear.
    assert len(ff(10, (20, 30), normalise=False).layers) == 5


def test_rnn():
    for final_dense, gru, nn in [
        (True, False, rnn(10, (20, 30), normalise=True, gru=False, final_dense=True)),
        (False, True, rnn(10, (20, 30), normalise=True, gru=True, final_dense=False)),
    ]:
        vs = Vars(np.float32)
        nn.initialise(5, vs)
        x = B.randn(2, 3, 5)

        # Check number of weights and width.
        assert B.length(vs.get_vector()) == nn.num_weights(5)
        assert nn.width == 10

        # Test batch consistency.
        check_batch_consistency(nn, x)

        # Check composition.
        assert len(nn.layers) == 9 if final_dense else 7
        assert type(nn.layers[0]) == Recurrent
        assert type(nn.layers[0].cell) == GRU if gru else Elman
        assert nn.layers[0].width == 20
        assert type(nn.layers[1]) == Activation
        assert nn.layers[1].width == 20
        assert type(nn.layers[2]) == Normalise
        assert nn.layers[2].width == 20
        assert type(nn.layers[3]) == Recurrent
        assert type(nn.layers[3].cell) == GRU if gru else Elman
        assert nn.layers[3].width == 30
        assert type(nn.layers[4]) == Activation
        assert nn.layers[4].width == 30
        assert type(nn.layers[5]) == Normalise
        assert nn.layers[5].width == 30
        if final_dense:
            assert type(nn.layers[6]) == Linear
            assert nn.layers[6].width == 10
            assert type(nn.layers[7]) == Activation
            assert nn.layers[7].width == 10
            assert type(nn.layers[8]) == Linear
            assert nn.layers[8].width == 10
        else:
            assert type(nn.layers[6]) == Linear
            assert nn.layers[6].width == 10

    # Check that normalisation layers disappear.
    assert (
        len(rnn(10, (20, 30), normalise=False, gru=True, final_dense=False).layers) == 5
    )



import pandas as pd

from .data import data_path, split_df, date_to_decimal_year, resource

__all__ = ["load"]


def load(extended=False, nguyen=False):
    """Load the data.

    Args:
        extended (bool, optional): Also return years 2000 to 2010. Defaults
            to `False`.
        nguyen (bool, optional): Return the split from Nguyen & Bonilla (2014).

    Returns:
        tuple[:class:`pd.DataFrame`]: Three-tuple or four-tuple containing
            all the data, training data, test data, and potentially years
            2000 to 2010.
    """
    _fetch()

    rates = _load_csv(data_path("exchange", "exchange2007.csv"))

    if nguyen:
        # Return the split from Nguyen & Bonilla (2014).
        inds_cad = (49, 100)
        inds_jpy = (99, 150)
        inds_aud = (149, 200)
    else:
        # Return the split from Requeima et al. (2019).
        inds_cad = (49, 100)
        inds_jpy = (49, 150)
        inds_aud = (49, 200)

    # Split into train and test according to paper.
    test1, train = split_df(rates, inds_cad, ["USD/CAD"], iloc=True)
    test2, train = split_df(train, inds_jpy, ["USD/JPY"], iloc=True)
    test3, train = split_df(train, inds_aud, ["USD/AUD"], iloc=True)
    test = pd.concat([test1, test2, test3], axis=1)

    if extended:
        # Also return the rates from years 2000 to 2010.
        return (
            rates,
            train,
            test,
            pd.concat(
                [
                    _load_csv(data_path("exchange", f"exchange{year}.csv"))
                    for year in range(1990, 2015 + 1)
                ],
                axis=0,
            ),
        )

    return rates, train, test


def _load_csv(path):
    df = pd.read_csv(path, header=1)

    # Remove tail.
    df = df.iloc[:-1, :]

    # Extract and invert rates.
    rates = 1 / df.iloc[:, 3:]
    rates.columns = [
        "{}/{}".format(*reversed(name.split("/"))) for name in rates.columns
    ]

    # Set index to years.
    years = [date_to_decimal_year(x, "%Y/%m/%d") for x in df["YYYY/MM/DD"]]
    rates["year"] = years
    rates = rates.set_index("year")

    # Reorder according to paper.
    last_names = ["USD/XPT", "USD/CAD", "USD/JPY", "USD/AUD"]
    names = sorted(set(rates.columns) - set(last_names)) + last_names
    rates = rates[names]

    return rates


def _fetch():
    for year in range(1990, 2015 + 1):
        resource(
            target=data_path("exchange", f"exchange{year}.csv"),
            url="http://fx.sauder.ubc.ca/cgi/fxdata",
            post=True,
            data={
                "b": "USD",
                "c": ["X12", "XC3"],
                "rd": "",
                "fd": "1",
                "fm": "1",
                "fy": str(year),
                "ld": "31",
                "lm": "12",
                "ly": str(year),
                "y": "daily",
                "q": "volume",
                "f": "csv",
                "o": "",
            },
        )



import numpy as np
import pandas as pd

from .data import split_df

__all__ = ["load"]


def load():
    x = np.linspace(0, 6 * np.pi, 100)
    y = np.stack((np.sin(x), 0.1 * x, 2 * np.sin(x), 0.1 * x + 0.2 * np.sin(x)), axis=1)
    y = y + 0.2 * np.random.randn(*y.shape)

    # Construct data and reorder outputs.
    df = pd.DataFrame(
        y,
        index=pd.Index(x, name="time"),
        columns=["sin(x)", "x^2", "sin(x)^2", "x + sin(x)"],
    )

    # Select data from the paper.
    df_train, df_test = split_df(
        df, (4 * np.pi, 6 * np.pi + 1), ["sin(x)^2", "x + sin(x)"]
    )

    return df, df_train, df_test
